import asyncio
from decimal import Decimal
import logging
import os
from pathlib import Path
import platform
from typing import (
    cast,
    Any,
    Collection,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
)

import aiohttp
from aiodocker.docker import Docker, DockerContainer
from aiodocker.exceptions import DockerError
import async_timeout
import psutil

from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.utils import current_loop, nmget
from ai.backend.common.types import (
    DeviceName, DeviceId,
    DeviceModelInfo,
    SlotName, SlotTypes,
    MetricKey,
)
from .agent import Container
from .resources import (
    get_resource_spec_from_container,
)
from .. import __version__
from ..resources import (
    AbstractAllocMap, DeviceSlotInfo,
    DiscretePropertyAllocMap,
    AbstractComputeDevice,
    AbstractComputePlugin,
)
from ..stats import (
    StatContext, NodeMeasurement, ContainerMeasurement,
    StatModes, MetricTypes, Measurement,
)
from ..utils import closing_async, read_sysfs
from ..vendor.linux import libnuma

log = BraceStyleAdapter(logging.getLogger(__name__))


async def fetch_api_stats(container: DockerContainer) -> Optional[Dict[str, Any]]:
    short_cid = container._id[:7]
    try:
        ret = await container.stats(stream=False)  # TODO: cache
    except RuntimeError as e:
        msg = str(e.args[0]).lower()
        if 'event loop is closed' in msg or 'session is closed' in msg:
            return None
        raise
    except (DockerError, aiohttp.ClientError) as e:
        log.error(
            'cannot read stats (cid:{}): client error: {!r}.',
            short_cid, e,
        )
        return None
    else:
        # aiodocker 0.16 or later returns a list of dict, even when not streaming.
        if isinstance(ret, list):
            if not ret:
                # The API may return an empty result upon container termination.
                return None
            ret = ret[0]
        # The API may return an invalid or empty result upon container termination.
        if ret is None or not isinstance(ret, dict):
            log.warning(
                'cannot read stats (cid:{}): got an empty result: {}',
                short_cid, ret,
            )
            return None
        if (
            ret['read'].startswith('0001-01-01') or
            ret['preread'].startswith('0001-01-01')
        ):
            return None
        return ret


# Pseudo-plugins for intrinsic devices (CPU and the main memory)

class CPUDevice(AbstractComputeDevice):
    pass


class CPUPlugin(AbstractComputePlugin):
    """
    Represents the CPU.
    """

    config_watch_enabled = False

    key = DeviceName('cpu')
    slot_types = [
        (SlotName('cpu'), SlotTypes.COUNT),
    ]

    async def init(self, context: Any = None) -> None:
        pass

    async def cleanup(self) -> None:
        pass

    async def update_plugin_config(self, new_plugin_config: Mapping[str, Any]) -> None:
        pass

    async def list_devices(self) -> Collection[CPUDevice]:
        cores = await libnuma.get_available_cores()
        overcommit_factor = int(os.environ.get('BACKEND_CPU_OVERCOMMIT_FACTOR', '1'))
        assert 1 <= overcommit_factor <= 10
        return [
            CPUDevice(
                device_id=DeviceId(str(core_idx)),
                hw_location='root',
                numa_node=libnuma.node_of_cpu(core_idx),
                memory_size=0,
                processing_units=1 * overcommit_factor,
            )
            for core_idx in sorted(cores)
        ]

    async def available_slots(self) -> Mapping[SlotName, Decimal]:
        devices = await self.list_devices()
        return {
            SlotName('cpu'): Decimal(sum(dev.processing_units for dev in devices)),
        }

    def get_version(self) -> str:
        return __version__

    async def extra_info(self) -> Mapping[str, str]:
        return {
            'agent_version': __version__,
            'machine': platform.machine(),
            'os_type': platform.system(),
        }

    async def gather_node_measures(self, ctx: StatContext) -> Sequence[NodeMeasurement]:
        _cstat = psutil.cpu_times(True)
        q = Decimal('0.000')
        total_cpu_used = cast(Decimal,
                              sum((Decimal(c.user + c.system) * 1000).quantize(q) for c in _cstat))
        now, raw_interval = ctx.update_timestamp('cpu-node')
        interval = Decimal(raw_interval * 1000).quantize(q)

        return [
            NodeMeasurement(
                MetricKey('cpu_util'),
                MetricTypes.UTILIZATION,
                unit_hint='msec',
                current_hook=lambda metric: metric.stats.diff,
                per_node=Measurement(total_cpu_used, interval),
                per_device={
                    DeviceId(str(idx)):
                    Measurement(
                        (Decimal(c.user + c.system) * 1000).quantize(q),
                        interval,
                    )
                    for idx, c in enumerate(_cstat)
                },
            ),
        ]

    async def gather_container_measures(
        self,
        ctx: StatContext,
        container_ids: Sequence[str],
    ) -> Sequence[ContainerMeasurement]:

        async def sysfs_impl(container_id):
            cpu_prefix = f'/sys/fs/cgroup/cpuacct/docker/{container_id}/'
            try:
                cpu_used = read_sysfs(cpu_prefix + 'cpuacct.usage', int) / 1e6
            except IOError as e:
                log.warning('cannot read stats: sysfs unreadable for container {0}\n{1!r}',
                            container_id[:7], e)
                return None
            return cpu_used

        async def api_impl(container_id):
            async with closing_async(Docker()) as docker:
                container = DockerContainer(docker, id=container_id)
                try:
                    async with async_timeout.timeout(2.0):
                        ret = await fetch_api_stats(container)
                except asyncio.TimeoutError:
                    return None
                if ret is None:
                    return None
                cpu_used = nmget(ret, 'cpu_stats.cpu_usage.total_usage', 0) / 1e6
                return cpu_used

        if ctx.mode == StatModes.CGROUP:
            impl = sysfs_impl
        elif ctx.mode == StatModes.DOCKER:
            impl = api_impl
        else:
            raise RuntimeError("should not reach here")

        q = Decimal('0.000')
        per_container_cpu_used = {}
        tasks = []
        for cid in container_ids:
            tasks.append(asyncio.ensure_future(impl(cid)))
        results = await asyncio.gather(*tasks)
        for cid, cpu_used in zip(container_ids, results):
            if cpu_used is None:
                continue
            per_container_cpu_used[cid] = Measurement(Decimal(cpu_used).quantize(q))
        return [
            ContainerMeasurement(
                MetricKey('cpu_util'),
                MetricTypes.UTILIZATION,
                unit_hint='percent',
                current_hook=lambda metric: metric.stats.rate,
                stats_filter=frozenset({'avg', 'max'}),
                per_container=per_container_cpu_used,
            ),
            ContainerMeasurement(
                MetricKey('cpu_used'),
                MetricTypes.USAGE,
                unit_hint='msec',
                per_container=per_container_cpu_used.copy(),
            ),
        ]

    async def create_alloc_map(self) -> AbstractAllocMap:
        devices = await self.list_devices()
        return DiscretePropertyAllocMap(
            device_slots={
                dev.device_id:
                    DeviceSlotInfo(SlotTypes.COUNT, SlotName('cpu'), Decimal(dev.processing_units))
                for dev in devices
            },
        )

    async def get_hooks(self, distro: str, arch: str) -> Sequence[Path]:
        # TODO: move the sysconf hook in libbaihook.so here
        return []

    async def generate_docker_args(
        self,
        docker: Docker,
        device_alloc,
    ) -> Mapping[str, Any]:
        cores = [*map(int, device_alloc['cpu'].keys())]
        sorted_core_ids = [*map(str, sorted(cores))]
        return {
            'HostConfig': {
                'CpuPeriod': 100_000,  # docker default
                'CpuQuota': int(100_000 * len(cores)),
                'Cpus': ','.join(sorted_core_ids),
                'CpusetCpus': ','.join(sorted_core_ids),
                # 'CpusetMems': f'{resource_spec.numa_node}',
            },
        }

    async def restore_from_container(
        self,
        container: Container,
        alloc_map: AbstractAllocMap,
    ) -> None:
        assert isinstance(alloc_map, DiscretePropertyAllocMap)
        # Docker does not return the original cpuset.... :(
        # We need to read our own records.
        resource_spec = await get_resource_spec_from_container(container.backend_obj)
        if resource_spec is None:
            return
        alloc_map.apply_allocation({
            SlotName('cpu'):
                resource_spec.allocations[DeviceName('cpu')][SlotName('cpu')],
        })

    async def get_attached_devices(
        self,
        device_alloc: Mapping[SlotName,
        Mapping[DeviceId, Decimal]],
    ) -> Sequence[DeviceModelInfo]:
        device_ids = [*device_alloc[SlotName('cpu')].keys()]
        available_devices = await self.list_devices()
        attached_devices: List[DeviceModelInfo] = []
        for device in available_devices:
            if device.device_id in device_ids:
                attached_devices.append({
                    'device_id': device.device_id,
                    'model_name': '',
                    'data': {'cores': len(device_ids)},
                })
        return attached_devices


class MemoryDevice(AbstractComputeDevice):
    pass


class MemoryPlugin(AbstractComputePlugin):
    """
    Represents the main memory.

    When collecting statistics, it also measures network and I/O usage
    in addition to the memory usage.
    """

    config_watch_enabled = False

    key = DeviceName('mem')
    slot_types = [
        (SlotName('mem'), SlotTypes.BYTES),
    ]

    async def init(self, context: Any = None) -> None:
        pass

    async def cleanup(self) -> None:
        pass

    async def update_plugin_config(self, new_plugin_config: Mapping[str, Any]) -> None:
        pass

    async def list_devices(self) -> Collection[MemoryDevice]:
        # TODO: support NUMA?
        memory_size = psutil.virtual_memory().total
        overcommit_factor = int(os.environ.get('BACKEND_MEM_OVERCOMMIT_FACTOR', '1'))
        return [MemoryDevice(
            device_id=DeviceId('root'),
            hw_location='root',
            numa_node=0,
            memory_size=overcommit_factor * memory_size,
            processing_units=0,
        )]

    async def available_slots(self) -> Mapping[SlotName, Decimal]:
        devices = await self.list_devices()
        return {
            SlotName('mem'): Decimal(sum(dev.memory_size for dev in devices)),
        }

    def get_version(self) -> str:
        return __version__

    async def extra_info(self) -> Mapping[str, str]:
        return {}

    async def gather_node_measures(self, ctx: StatContext) -> Sequence[NodeMeasurement]:
        _mstat = psutil.virtual_memory()
        total_mem_used_bytes = Decimal(_mstat.total - _mstat.available)
        total_mem_capacity_bytes = Decimal(_mstat.total)
        _nstat = psutil.net_io_counters()
        net_rx_bytes = _nstat.bytes_recv
        net_tx_bytes = _nstat.bytes_sent

        def get_disk_stat():
            pruned_disk_types = frozenset(['squashfs', 'vfat', 'tmpfs'])
            total_disk_usage = Decimal(0)
            total_disk_capacity = Decimal(0)
            per_disk_stat = {}
            for disk_info in psutil.disk_partitions():
                if disk_info.fstype not in pruned_disk_types:
                    dstat = os.statvfs(disk_info.mountpoint)
                    disk_usage = Decimal(dstat.f_frsize * (dstat.f_blocks - dstat.f_bavail))
                    disk_capacity = Decimal(dstat.f_frsize * dstat.f_blocks)
                    per_disk_stat[disk_info.device] = Measurement(disk_usage, disk_capacity)
                    total_disk_usage += disk_usage
                    total_disk_capacity += disk_capacity
            return total_disk_usage, total_disk_capacity, per_disk_stat

        loop = current_loop()
        total_disk_usage, total_disk_capacity, per_disk_stat = \
            await loop.run_in_executor(None, get_disk_stat)
        return [
            NodeMeasurement(
                MetricKey('mem'),
                MetricTypes.USAGE,
                unit_hint='bytes',
                stats_filter=frozenset({'max'}),
                per_node=Measurement(total_mem_used_bytes, total_mem_capacity_bytes),
                per_device={DeviceId('root'):
                            Measurement(total_mem_used_bytes,
                                        total_mem_capacity_bytes)},
            ),
            NodeMeasurement(
                MetricKey('disk'),
                MetricTypes.USAGE,
                unit_hint='bytes',
                per_node=Measurement(total_disk_usage, total_disk_capacity),
                per_device=per_disk_stat,
            ),
            NodeMeasurement(
                MetricKey('net_rx'),
                MetricTypes.RATE,
                unit_hint='bps',
                current_hook=lambda metric: metric.stats.rate,
                per_node=Measurement(Decimal(net_rx_bytes)),
                per_device={DeviceId('node'): Measurement(Decimal(net_rx_bytes))},
            ),
            NodeMeasurement(
                MetricKey('net_tx'),
                MetricTypes.RATE,
                unit_hint='bps',
                current_hook=lambda metric: metric.stats.rate,
                per_node=Measurement(Decimal(net_tx_bytes)),
                per_device={DeviceId('node'): Measurement(Decimal(net_tx_bytes))},
            ),
        ]

    async def gather_container_measures(self, ctx: StatContext, container_ids: Sequence[str]) \
            -> Sequence[ContainerMeasurement]:

        def get_scratch_size(container_id: str) -> int:
            # Temporarily disabled as this function incurs too much delay with
            # a large number of files in scratch dirs, causing indefinite accumulation of
            # stat collector tasks and slowing down everything.
            return 0
            # for kernel_id, info in ctx.agent.kernel_registry.items():
            #     if info['container_id'] == container_id:
            #         break
            # else:
            #     return 0
            # work_dir = ctx.agent.local_config['container']['scratch-root'] / str(kernel_id) / 'work'
            # total_size = 0
            # for path in work_dir.rglob('*'):
            #     if path.is_symlink():
            #         total_size += path.lstat().st_size
            #     elif path.is_file():
            #         total_size += path.stat().st_size
            # return total_size

        async def sysfs_impl(container_id):
            mem_prefix = f'/sys/fs/cgroup/memory/docker/{container_id}/'
            io_prefix = f'/sys/fs/cgroup/blkio/docker/{container_id}/'
            try:
                mem_cur_bytes = read_sysfs(mem_prefix + 'memory.usage_in_bytes', int)
                io_stats = Path(io_prefix + 'blkio.throttle.io_service_bytes').read_text()
                # example data:
                #   8:0 Read 13918208
                #   8:0 Write 0
                #   8:0 Sync 0
                #   8:0 Async 13918208
                #   8:0 Total 13918208
                #   Total 13918208
                io_read_bytes = 0
                io_write_bytes = 0
                for line in io_stats.splitlines():
                    if line.startswith('Total '):
                        continue
                    dev, op, nbytes = line.strip().split()
                    if op == 'Read':
                        io_read_bytes += int(nbytes)
                    elif op == 'Write':
                        io_write_bytes += int(nbytes)
            except IOError as e:
                log.warning('cannot read stats: sysfs unreadable for container {0}\n{1!r}',
                            container_id[:7], e)
                return None
            loop = current_loop()
            scratch_sz = await loop.run_in_executor(
                None, get_scratch_size, container_id)
            return mem_cur_bytes, io_read_bytes, io_write_bytes, scratch_sz

        async def api_impl(container_id):
            async with closing_async(Docker()) as docker:
                container = DockerContainer(docker, id=container_id)
                try:
                    async with async_timeout.timeout(2.0):
                        ret = await fetch_api_stats(container)
                except asyncio.TimeoutError:
                    return None
                if ret is None:
                    return None
                mem_cur_bytes = nmget(ret, 'memory_stats.usage', 0)
                io_read_bytes = 0
                io_write_bytes = 0
                for item in nmget(ret, 'blkio_stats.io_service_bytes_recursive', []):
                    if item['op'] == 'Read':
                        io_read_bytes += item['value']
                    elif item['op'] == 'Write':
                        io_write_bytes += item['value']
                loop = current_loop()
                scratch_sz = await loop.run_in_executor(
                    None, get_scratch_size, container_id)
                return mem_cur_bytes, io_read_bytes, io_write_bytes, scratch_sz

        if ctx.mode == StatModes.CGROUP:
            impl = sysfs_impl
        elif ctx.mode == StatModes.DOCKER:
            impl = api_impl
        else:
            raise RuntimeError("should not reach here")

        per_container_mem_used_bytes = {}
        per_container_io_read_bytes = {}
        per_container_io_write_bytes = {}
        per_container_io_scratch_size = {}
        tasks = []
        for cid in container_ids:
            tasks.append(asyncio.ensure_future(impl(cid)))
        results = await asyncio.gather(*tasks)
        for cid, result in zip(container_ids, results):
            if result is None:
                continue
            per_container_mem_used_bytes[cid] = Measurement(
                Decimal(result[0]))
            per_container_io_read_bytes[cid] = Measurement(
                Decimal(result[1]))
            per_container_io_write_bytes[cid] = Measurement(
                Decimal(result[2]))
            per_container_io_scratch_size[cid] = Measurement(
                Decimal(result[3]))
        return [
            ContainerMeasurement(
                MetricKey('mem'),
                MetricTypes.USAGE,
                unit_hint='bytes',
                stats_filter=frozenset({'max'}),
                per_container=per_container_mem_used_bytes,
            ),
            ContainerMeasurement(
                MetricKey('io_read'),
                MetricTypes.USAGE,
                unit_hint='bytes',
                stats_filter=frozenset({'rate'}),
                per_container=per_container_io_read_bytes,
            ),
            ContainerMeasurement(
                MetricKey('io_write'),
                MetricTypes.USAGE,
                unit_hint='bytes',
                stats_filter=frozenset({'rate'}),
                per_container=per_container_io_write_bytes,
            ),
            ContainerMeasurement(
                MetricKey('io_scratch_size'),
                MetricTypes.USAGE,
                unit_hint='bytes',
                stats_filter=frozenset({'max'}),
                per_container=per_container_io_scratch_size,
            ),
        ]

    async def create_alloc_map(self) -> AbstractAllocMap:
        devices = await self.list_devices()
        return DiscretePropertyAllocMap(
            device_slots={
                dev.device_id:
                    DeviceSlotInfo(SlotTypes.BYTES, SlotName('mem'), Decimal(dev.memory_size))
                for dev in devices
            },
        )

    async def get_hooks(self, distro: str, arch: str) -> Sequence[Path]:
        return []

    async def generate_docker_args(
        self,
        docker: Docker,
        device_alloc,
    ) -> Mapping[str, Any]:
        memory = sum(device_alloc['mem'].values())
        return {
            'HostConfig': {
                'MemorySwap': int(memory),  # prevent using swap!
                'Memory': int(memory),
            },
        }

    async def restore_from_container(
        self,
        container: Container,
        alloc_map: AbstractAllocMap,
    ) -> None:
        assert isinstance(alloc_map, DiscretePropertyAllocMap)
        memory_limit = container.backend_obj['HostConfig']['Memory']
        alloc_map.apply_allocation({
            SlotName('mem'): {DeviceId('root'): memory_limit},
        })

    async def get_attached_devices(
        self,
        device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]],
    ) -> Sequence[DeviceModelInfo]:
        device_ids = [*device_alloc[SlotName('mem')].keys()]
        available_devices = await self.list_devices()
        attached_devices: List[DeviceModelInfo] = []
        for device in available_devices:
            if device.device_id in device_ids:
                attached_devices.append({
                    'device_id': device.device_id,
                    'model_name': '',
                    'data': {},
                })
        return attached_devices



import asyncio
from decimal import Decimal
import hashlib
import io
import ipaddress
import json
import logging
from pathlib import Path
import platform
import re
from typing import (
    Any,
    AsyncContextManager,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Protocol,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)
from typing_extensions import Final
from uuid import UUID

import aiodocker
from aiodocker.docker import DockerContainer
import netifaces
import trafaret as t

from ai.backend.common import identity
from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import (
    PID, HostPID, ContainerPID, KernelId,
)
from ai.backend.common.utils import current_loop

log = BraceStyleAdapter(logging.getLogger('ai.backend.agent.utils'))

IPNetwork = Union[ipaddress.IPv4Network, ipaddress.IPv6Network]
IPAddress = Union[ipaddress.IPv4Address, ipaddress.IPv6Address]

InOtherContainerPID: Final = ContainerPID(PID(-2))
NotContainerPID: Final = ContainerPID(PID(-1))
NotHostPID: Final = HostPID(PID(-1))


class SupportsAsyncClose(Protocol):
    async def close(self) -> None:
        ...


_SupportsAsyncCloseT = TypeVar('_SupportsAsyncCloseT', bound=SupportsAsyncClose)


class closing_async(AsyncContextManager[_SupportsAsyncCloseT]):
    """
    contextlib.closing calls close(), and aiotools.aclosing() calls aclose().
    This context manager calls close() as a coroutine.
    """

    def __init__(self, obj: _SupportsAsyncCloseT) -> None:
        self.obj = obj

    async def __aenter__(self) -> _SupportsAsyncCloseT:
        return self.obj

    async def __aexit__(self, *exc_info) -> None:
        await self.obj.close()


def generate_local_instance_id(hint: str) -> str:
    return hashlib.md5(hint.encode('utf-8')).hexdigest()[:12]


def get_arch_name() -> str:
    ret = platform.machine().lower()
    aliases = {
        "arm64": "aarch64",  # macOS with LLVM
        "amd64": "x86_64",   # Windows/Linux
        "x64": "x86_64",     # Windows
        "x32": "x86",        # Windows
        "i686": "x86",       # Windows
    }
    return aliases.get(ret, ret)


def update_nested_dict(dest: MutableMapping, additions: Mapping) -> None:
    for k, v in additions.items():
        if k not in dest:
            dest[k] = v
        else:
            if isinstance(dest[k], MutableMapping):
                assert isinstance(v, MutableMapping)
                update_nested_dict(dest[k], v)
            elif isinstance(dest[k], List):
                assert isinstance(v, List)
                dest[k].extend(v)
            else:
                dest[k] = v


def numeric_list(s: str) -> List[int]:
    return [int(p) for p in s.split()]


def remove_exponent(num: Decimal) -> Decimal:
    return num.quantize(Decimal(1)) if num == num.to_integral() else num.normalize()


@overload
def read_sysfs(path: Union[str, Path], type_: Type[bool], default: bool) -> bool:
    ...


@overload
def read_sysfs(path: Union[str, Path], type_: Type[int], default: int) -> int:
    ...


@overload
def read_sysfs(path: Union[str, Path], type_: Type[float], default: float) -> float:
    ...


@overload
def read_sysfs(path: Union[str, Path], type_: Type[str], default: str) -> str:
    ...


def read_sysfs(path: Union[str, Path], type_: Type[Any], default: Any = None) -> Any:
    def_vals: Mapping[Any, Any] = {
        bool: False,
        int: 0,
        float: 0.0,
        str: '',
    }
    if type_ not in def_vals:
        raise TypeError('unsupported conversion type from sysfs content')
    if default is None:
        default = def_vals[type_]
    try:
        raw_str = Path(path).read_text().strip()
        if type_ is bool:
            return t.ToBool().check(raw_str)
        else:
            return type_(raw_str)
    except IOError:
        return default


async def read_tail(path: Path, nbytes: int) -> bytes:
    file_size = path.stat().st_size

    def _read_tail() -> bytes:
        with open(path, 'rb') as f:
            f.seek(max(file_size - nbytes, 0), io.SEEK_SET)
            return f.read(nbytes)

    loop = current_loop()
    return await loop.run_in_executor(None, _read_tail)


async def get_kernel_id_from_container(val: Union[str, DockerContainer]) -> Optional[KernelId]:
    if isinstance(val, DockerContainer):
        if 'Name' not in val._container:
            await val.show()
        name = val['Name']
    elif isinstance(val, str):
        name = val
    name = name.lstrip('/')
    if not name.startswith('kernel.'):
        return None
    try:
        return KernelId(UUID(name.rsplit('.', 2)[-1]))
    except (IndexError, ValueError):
        return None


async def get_subnet_ip(etcd: AsyncEtcd, network: str, fallback_addr: str = '0.0.0.0') -> str:
    raw_subnet = await etcd.get(f'config/network/subnet/{network}')
    if raw_subnet is None:
        addr = fallback_addr
    else:
        subnet = ipaddress.ip_network(raw_subnet)
        if subnet.prefixlen == 0:
            addr = fallback_addr
        else:
            local_ipaddrs = [*identity.fetch_local_ipaddrs(subnet)]
            log.debug('get_subnet_ip(): subnet {} candidates: {}',
                      subnet, local_ipaddrs)
            if local_ipaddrs:
                addr = str(local_ipaddrs[0])
            else:
                addr = fallback_addr
    return addr


async def host_pid_to_container_pid(container_id: str, host_pid: HostPID) -> ContainerPID:
    kernel_ver = Path('/proc/version').read_text()
    if m := re.match(r'Linux version (\d+)\.(\d+)\..*', kernel_ver):  # noqa
        kernel_ver_tuple: Tuple[str, str] = m.groups()  # type: ignore
        if kernel_ver_tuple < ('4', '1'):
            # TODO: this should be deprecated when the minimun supported Linux kernel will be 4.1.
            #
            # In CentOs 7, NSPid is not accesible since it is supported from Linux kernel >=4.1.
            # We provide alternative, although messy, way for older Linux kernels. Below describes
            # the logic briefly:
            #   * Obtain information on all the processes inside the target container,
            #     which contains host PID, by docker top API (containers/<container-id>/top).
            #     - Get the COMMAND of the target process (by using host_pid).
            #     - Filter host processes which have the exact same COMMAND.
            #   * Obtain information on all the processes inside the target container,
            #     which contains container PID, by executing "ps -aux" command from inside the container.
            #     - Filter container processes which have the exact same COMMAND.
            #   * Get the index of the target process from the host process table.
            #   * Use the index to get the target process from the container process table, and get PID.
            #     - Since docker top and ps -aux both displays processes in the order of PID, we
            #       can safely assume that the order of the processes from both tables are the same.
            #
            # Example host and container process table:
            #
            # [
            #   ['devops', '15454', '12942', '99', '15:36', 'pts/1', '00:00:08', 'python mnist.py'],
            #   ... (processes with the same COMMAND)
            # ]
            #
            # [
            #   ['work', '227', '121', '4.6', '22408680', '1525428', 'pts/1', 'Rl+', '06:36', '0:08',
            #    'python', 'mnist.py'],
            #   ... (processes with the same COMMAND)
            # ]
            try:
                docker = aiodocker.Docker()
                # Get process table from host (docker top information). Filter processes which have
                # exactly the same COMMAND as with target host process.
                result = await docker._query_json(f'containers/{container_id}/top', method='GET')
                procs = result['Processes']
                cmd = list(filter(lambda x: str(host_pid) == x[1], procs))[0][7]
                host_table = list(filter(lambda x: cmd == x[7], procs))

                # Get process table from inside container (execute 'ps -aux' command from container).
                # Filter processes which have exactly the same COMMAND like above.
                result = await docker._query_json(
                    f'containers/{container_id}/exec',
                    method='POST',
                    data={
                        'AttachStdin': False,
                        'AttachStdout': True,
                        'AttachStderr': True,
                        'Cmd': ['ps', '-aux'],
                    },
                )
                exec_id = result['Id']
                async with docker._query(
                    f'exec/{exec_id}/start',
                    method='POST',
                    headers={'content-type': 'application/json'},
                    data=json.dumps({
                        'Stream': False,  # get response immediately
                        'Detach': False,
                        'Tty': False,
                    }),
                ) as resp:
                    result = await resp.read()
                    result = result.decode('latin-1').split('\n')
                result = list(map(lambda x: x.split(), result))
                head = result[0]
                procs = result[1:]
                pid_idx, cmd_idx = head.index('PID'), head.index('COMMAND')
                container_table = list(
                    filter(lambda x: cmd == ' '.join(x[cmd_idx:]) if x else False, procs),
                )

                # When there are multiple processes which have the same COMMAND, just get the index of
                # the target host process and apply it with the container table. Since ps and docker top
                # both displays processes ordered by PID, we can expect those two tables have same
                # order of processes.
                process_idx = None
                for idx, p in enumerate(host_table):
                    if str(host_pid) == p[1]:
                        process_idx = idx
                        break
                else:
                    raise IndexError
                container_pid = ContainerPID(container_table[process_idx][pid_idx])
                log.debug('host pid {} is mapped to container pid {}', host_pid, container_pid)
                return ContainerPID(PID(int(container_pid)))
            except asyncio.CancelledError:
                raise
            except (IndexError, KeyError, aiodocker.exceptions.DockerError):
                return NotContainerPID
            finally:
                await docker.close()

    try:
        for p in Path('/sys/fs/cgroup/pids/docker').iterdir():
            if not p.is_dir():
                continue
            tasks_path = p / 'tasks'
            cgtasks = [*map(int, tasks_path.read_text().splitlines())]
            if host_pid not in cgtasks:
                continue
            if p.name == container_id:
                proc_path = Path(f'/proc/{host_pid}/status')
                proc_status = {k: v for k, v
                               in map(lambda l: l.split(':\t'),
                                      proc_path.read_text().splitlines())}
                nspids = [*map(lambda pid: ContainerPID(PID(int(pid))), proc_status['NSpid'].split())]
                return nspids[1]
            return InOtherContainerPID
        return NotContainerPID
    except (ValueError, KeyError, IOError):
        return NotContainerPID


async def container_pid_to_host_pid(container_id: str, container_pid: ContainerPID) -> HostPID:
    # TODO: implement
    return NotHostPID


def fetch_local_ipaddrs(cidr: IPNetwork) -> Iterable[IPAddress]:
    ifnames = netifaces.interfaces()
    proto = netifaces.AF_INET if cidr.version == 4 else netifaces.AF_INET6
    for ifname in ifnames:
        addrs = netifaces.ifaddresses(ifname).get(proto, None)
        if addrs is None:
            continue
        for entry in addrs:
            addr = ipaddress.ip_address(entry['addr'])
            if addr in cidr:
                yield addr



import os
import socket
import sys

input_host = '127.0.0.1'
input_port = 65000

batch_enabled = int(os.environ.get('_BACKEND_BATCH_MODE', '0'))
if batch_enabled:
    # Since latest Python 2 has `builtins`and `input`,
    # we cannot detect Python 2 with the existence of them.
    if sys.version_info.major > 2:
        import builtins

        def _input(prompt=''):
            sys.stdout.write(prompt)
            sys.stdout.flush()
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                try:
                    sock.connect((input_host, input_port))
                    userdata = sock.recv(1024)
                except ConnectionRefusedError:
                    userdata = b'<user-input-unavailable>'
            return userdata.decode()
        builtins._input = input  # type: ignore
        builtins.input = _input
    else:
        # __builtins__ is an alias dict for __builtin__ in modules other than __main__.
        # Thus, we have to explicitly import __builtin__ module in Python 2.
        import __builtin__
        builtins = __builtin__

        def _raw_input(prompt=''):
            sys.stdout.write(prompt)
            sys.stdout.flush()
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.connect((input_host, input_port))
                userdata = sock.recv(1024)
            except socket.error:
                userdata = b'<user-input-unavailable>'
            finally:
                sock.close()
            return userdata.decode()
        builtins._raw_input = builtins.raw_input  # type: ignore
        builtins.raw_input = _raw_input           # type: ignore



import asyncio
import gzip
import logging
from pathlib import Path
import pkg_resources
import subprocess
from typing import Any, BinaryIO, Mapping, Tuple, cast

from aiodocker.docker import Docker
from aiodocker.exceptions import DockerError

from ai.backend.common.logging import BraceStyleAdapter

from ..exception import InitializationError
from ..utils import closing_async, get_arch_name, update_nested_dict

log = BraceStyleAdapter(logging.getLogger(__name__))


class PersistentServiceContainer:

    def __init__(
        self,
        image_ref: str,
        container_config: Mapping[str, Any],
        *,
        name: str = None,
    ) -> None:
        self.image_ref = image_ref
        arch = get_arch_name()
        default_container_name = image_ref.split(':')[0].rsplit('/', maxsplit=1)[-1]
        if name is None:
            self.container_name = default_container_name
        else:
            self.container_name = name
        self.container_config = container_config
        self.img_version = int(Path(pkg_resources.resource_filename(
            'ai.backend.agent.docker',
            f'{default_container_name}.version.txt',
        )).read_text())
        self.img_path = Path(pkg_resources.resource_filename(
            'ai.backend.agent.docker',
            f'{default_container_name}.img.{arch}.tar.gz',
        ))

    async def get_container_version_and_status(self) -> Tuple[int, bool]:
        async with closing_async(Docker()) as docker:
            try:
                c = docker.containers.container(self.container_name)
                await c.show()
            except DockerError as e:
                if e.status == 404:
                    return 0, False
                else:
                    raise
        if c['Config'].get('Labels', {}).get('ai.backend.system', '0') != '1':
            raise RuntimeError(
                f"An existing container named \"{c['Name'].lstrip('/')}\" is not a system container "
                f"spawned by Backend.AI. Please check and remove it.")
        return (
            int(c['Config'].get('Labels', {}).get('ai.backend.version', '0')),
            c['State']['Status'].lower() == 'running',
        )

    async def get_image_version(self) -> int:
        async with closing_async(Docker()) as docker:
            try:
                img = await docker.images.inspect(self.image_ref)
            except DockerError as e:
                if e.status == 404:
                    return 0
                else:
                    raise
        return int((img['Config'].get('Labels') or {}).get('ai.backend.version', '0'))

    async def ensure_running_latest(self) -> None:
        image_version = await self.get_image_version()
        if image_version == 0:
            log.info("PersistentServiceContainer({}): installing...", self.image_ref)
            await self.install_latest()
        elif image_version < self.img_version:
            log.info("PersistentServiceContainer({}): upgrading (v{} -> v{})",
                     self.image_ref, image_version, self.img_version)
            await self.install_latest()
        container_version, is_running = await self.get_container_version_and_status()
        if container_version == 0 or image_version != container_version or not is_running:
            log.info("PersistentServiceContainer({}): recreating...", self.image_ref)
            await self.recreate()
        if not is_running:
            log.info("PersistentServiceContainer({}): starting...", self.image_ref)
            await self.start()

    async def install_latest(self) -> None:
        with gzip.open(self.img_path, 'rb') as reader:
            proc = await asyncio.create_subprocess_exec(
                *['docker', 'load'],
                stdin=cast(BinaryIO, reader),
                stdout=subprocess.DEVNULL,
                stderr=subprocess.PIPE,
            )
            if (await proc.wait() != 0):
                stderr = b'(unavailable)'
                if proc.stderr is not None:
                    stderr = await proc.stderr.read()
                raise RuntimeError(
                    'loading the image has failed!',
                    self.image_ref, proc.returncode, stderr,
                )

    async def recreate(self) -> None:
        async with closing_async(Docker()) as docker:
            try:
                c = docker.containers.container(self.container_name)
                await c.stop()
                await c.delete(force=True)
            except DockerError as e:
                if e.status == 409 and 'is not running' in e.message:
                    pass
                elif e.status == 404:
                    pass
                else:
                    raise
            container_config = {
                'Image': self.image_ref,
                'Tty': True,
                'Privileged': False,
                'AttachStdin': False,
                'AttachStdout': False,
                'AttachStderr': False,
                'HostConfig': {
                    'Init': True,
                    'RestartPolicy': {
                        'Name': 'unless-stopped',  # make it persistent
                        'MaximumRetryCount': 0,
                    },
                },
            }
            update_nested_dict(container_config, self.container_config)
            try:
                await docker.containers.create(config=container_config, name=self.container_name)
            except DockerError as e:
                err_msg = e.args[1].get("message", "")
                if (
                    e.args[0] == 400 and
                    'bind source path does not exist' in err_msg and
                    '/tmp/backend.ai/ipc' in err_msg
                ):
                    raise InitializationError(
                        f"Could not create persistent service container '{self.container_name}' "
                        f"because it cannot access /tmp/backend.ai/ipc directory. "
                        f"This may occur when Docker is installed with Snap or the agent is configured "
                        f"to use a private tmp directory. "
                        f"To resolve, explicitly configure the 'ipc-base-path' option in agent.toml to "
                        f"indicate a directory under $HOME or a non-virtualized directory.",
                    )
                else:
                    raise

    async def start(self) -> None:
        async with closing_async(Docker()) as docker:
            c = docker.containers.container(self.container_name)
            await c.start()



from six.moves import builtins

from ..types import MediaRecord
from .encoding import encode_commands
from .turtle import Turtle
from .color import Colors


_canvas_id_counter = 0


class DrawingObject:

    def __init__(self, canvas, id_, args):
        self._canvas = canvas
        self._id = id_
        self._type = args[0]

    def set_x(self, x):
        if self._type in (u'rect', u'circle', u'triangle'):
            self._canvas._cmd_history.append((
                self._canvas._id, u'update', self._id,
                u'x', x))

    def set_y(self, y):
        if self._type in (u'rect', u'circle', u'triangle'):
            self._canvas._cmd_history.append((
                self._canvas._id, u'update', self._id,
                u'y', y))

    def set_x1(self, x):
        if self._type == u'line':
            self._canvas._cmd_history.append((
                self._canvas._id, u'update', self._id,
                u'x1', x))

    def set_y1(self, y):
        if self._type == u'line':
            self._canvas._cmd_history.append((
                self._canvas._id, u'update', self._id,
                u'y1', y))

    def set_x2(self, x):
        if self._type == u'line':
            self._canvas._cmd_history.append((
                self._canvas._id, u'update', self._id,
                u'x2', x))

    def set_y2(self, y):
        if self._type == u'line':
            self._canvas._cmd_history.append((
                self._canvas._id, u'update', self._id,
                u'y2', y))

    def set_radius(self, r):
        if self._type == u'circle':
            self._canvas._cmd_history.append((
                self._canvas._id, u'update', self._id,
                u'radius', r))

    def rotate(self, a):
        self._canvas._cmd_history.append((
            self._canvas._id, u'update', self._id,
            u'rotate', a))

    def set_angle(self, a):
        self._canvas._cmd_history.append((
            self._canvas._id, u'update', self._id,
            u'angle', a))

    def stroke(self, color):
        color = color.to_hex()
        if self._type == u'line':
            self._canvas._cmd_history.append((
                self._canvas._id, u'update', self._id,
                u'color', color))
        elif self._type == u'circle':
            self._canvas._cmd_history.append((
                self._canvas._id, u'update', self._id,
                u'border', color))
        elif self._type in (u'rect', u'triangle'):
            self._canvas._cmd_history.append((
                self._canvas._id, u'update', self._id,
                u'border', color))

    def fill(self, color):
        color = color.to_hex()
        if self._type == u'circle':
            self._canvas._cmd_history.append((
                self._canvas._id, u'update', self._id,
                u'fill', color))
        elif self._type in (u'rect', u'triangle'):
            self._canvas._cmd_history.append((
                self._canvas._id, u'update', self._id,
                u'fill', color))


class Canvas:

    def __init__(self, width, height, bgcolor=Colors.White, fgcolor=Colors.Black):
        global _canvas_id_counter
        self._id = _canvas_id_counter
        _canvas_id_counter += 1
        self._cmd_history = []
        self._next_objid = 0
        self._cmd_history.append((self._id, u'canvas',
                                 width, height,
                                 bgcolor.to_hex(),
                                 fgcolor.to_hex()))
        self.width = width
        self.height = height
        self.bgcolor = bgcolor
        self.fgcolor = fgcolor

    def update(self):
        builtins._sorna_emit(MediaRecord(
            u'application/x-sorna-drawing',
            encode_commands(self._cmd_history),
        ))
        self._cmd_history = []

    def show(self):  # alias
        self.update()

    def create_turtle(self):
        t = Turtle(self)
        return t

    def stop_animation(self):
        self._cmd_history.append((self._id, u'stop-anim'))

    def resume_animation(self):
        self._cmd_history.append((self._id, u'resume-anim'))

    def begin_group(self):
        self._cmd_history.append((self._id, u'begin-group'))

    def end_group(self):
        self._cmd_history.append((self._id, u'end-group'))

    def begin_fill(self, c):
        self._cmd_history.append((self._id, u'begin-fill', c.to_hex()))

    def end_fill(self):
        self._cmd_history.append((self._id, u'end-fill'))

    def background_color(self, c):
        self.bgcolor = c
        self._cmd_history.append((self._id, u'bgcolor', c.to_hex()))

    def stroke_color(self, c):
        self.fgcolor = c
        self._cmd_history.append((self._id, u'fgcolor', c.to_hex()))

    def line(self, x0, y0, x1, y1, color=None):
        if color is None:
            color = self.fgcolor
        args = (
            u'line', x0, y0, x1, y1,
            color.to_hex(),
        )
        self._cmd_history.append((self._id, u'obj', self._next_objid, args))
        obj = DrawingObject(self, self._next_objid, args)
        self._next_objid += 1
        return obj

    def circle(self, x, y, radius, border=None, fill=None, angle=0):
        if border is None:
            border = self.fgcolor
        if fill is None:
            fill = Colors.Transparent
        args = (
            u'circle', x, y, radius,
            border.to_hex(), fill.to_hex(), angle,
        )
        self._cmd_history.append((self._id, u'obj', self._next_objid, args))
        obj = DrawingObject(self, self._next_objid, args)
        self._next_objid += 1
        return obj

    def rectangle(self, left, top, width, height, border=None, fill=None, angle=0):
        if border is None:
            border = self.fgcolor
        if fill is None:
            fill = Colors.Transparent
        args = (
            u'rect', left, top, width, height,
            border.to_hex(), fill.to_hex(), angle,
        )
        self._cmd_history.append((self._id, u'obj', self._next_objid, args))
        obj = DrawingObject(self, self._next_objid, args)
        self._next_objid += 1
        return obj

    def triangle(self, left, top, width, height, border=None, fill=None, angle=0):
        if border is None:
            border = self.fgcolor
        if fill is None:
            fill = Colors.Transparent
        args = (
            u'triangle', left, top, width, height,
            border.to_hex(), fill.to_hex(), angle,
        )
        self._cmd_history.append((self._id, u'obj', self._next_objid, args))
        obj = DrawingObject(self, self._next_objid, args)
        self._next_objid += 1
        return obj


__all__ = [
    'Canvas',
]



import asyncio
import json
import logging
import os
from pathlib import Path
import re
import subprocess

# import pygit2
# from pygit2 import GIT_SORT_TOPOLOGICAL, GIT_SORT_REVERSE

from .. import BaseRunner, Terminal

log = logging.getLogger()

CHILD_ENV = {
    'TERM': 'xterm',
    'LANG': 'C.UTF-8',
    'SHELL': '/bin/bash',
    'USER': 'work',
    'HOME': '/home/work',
    'PATH': '/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin',
    'LD_PRELOAD': os.environ.get('LD_PRELOAD', '/home/backend.ai/libbaihook.so'),
}


class Runner(BaseRunner):

    log_prefix = 'shell-kernel'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.child_env.update(CHILD_ENV)

    async def init_with_loop(self):
        self.user_input_queue = asyncio.Queue()
        self.term = Terminal(
            '/bin/bash',
            self.stopped, self.outsock,
            auto_restart=True,
        )

        parser_show = self.term.subparsers.add_parser('show')
        parser_show.add_argument('target', choices=('graph',), default='graph')
        parser_show.add_argument('path', type=str)
        parser_show.set_defaults(func=self.do_show)

        await self.term.start()

    async def build_heuristic(self) -> int:
        raise NotImplementedError

    async def execute_heuristic(self) -> int:
        raise NotImplementedError

    async def query(self, code_text) -> int:
        return await self.term.handle_command(code_text)

    async def complete(self, data):
        return []

    async def interrupt(self):
        # subproc interrupt is already handled by BaseRunner
        pass

    async def shutdown(self):
        await self.term.shutdown()

    def do_show(self, args):
        if args.target == 'graph':
            commit_branch_table = {}
            commit_info = []

            if args.path in ['.', None]:
                current_dir = Path(f'/proc/{self.term.pid}/cwd').resolve()
            else:
                current_dir = Path(args.path).resolve()
            os.chdir(current_dir)

            # Create commit-branch matching table.
            tree_cmd = ['git', 'log', '--pretty=oneline', '--graph',
                        '--source', '--branches']
            run_res = subprocess.run(tree_cmd, stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE)
            stdout = run_res.stdout.decode('utf-8')
            stderr = run_res.stderr.decode('utf-8')
            prog = re.compile(r'([a-z0-9]+)\s+(\S+).*')
            if stderr:
                self.outsock.send_multipart([b'stderr', stderr.encode('utf-8')])
                return

            for line in stdout.split('\n'):
                r = prog.search(line)
                if r and hasattr(r, 'group') and r.group(1) and r.group(2):
                    oid = r.group(1)[:7]  # short oid
                    branch = r.group(2)
                    commit_branch_table[oid] = branch

            # Gather commit info w/ branch name.
            log_cmd = ['git', 'log', '--pretty=format:%h||%p||%s||%cn',
                       '--all', '--topo-order', '--reverse']
            run_res = subprocess.run(log_cmd, stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE)
            stdout = run_res.stdout.decode('utf-8')
            for gitlog in stdout.split('\n'):
                items = gitlog.split('||')
                oid = items[0]
                parent_ids = items[1].split(' ')
                message = items[2]
                author = items[3]
                branch = commit_branch_table.get(oid, None)
                parent_branches = [commit_branch_table.get(pid, None)
                                   for pid in parent_ids]
                info = dict(
                    oid=oid,
                    parent_ids=parent_ids,
                    author=author,
                    message=message,
                    branch=branch,
                    parent_branches=parent_branches,
                )
                commit_info.append(info)

            self.outsock.send_multipart([
                b'media',
                json.dumps({
                    'type': 'application/vnd.sorna.gitgraph',
                    'data': commit_info,
                }).encode('utf-8'),
            ])
        else:
            raise ValueError('Unsupported show target', args.target)

    async def start_service(self, service_info):
        return None, {}



import asyncio
import logging
import lzma
from pathlib import Path
import shutil
from ai.backend.agent.utils import get_arch_name
import pkg_resources
import re
import textwrap
from typing import (
    Any, Optional,
    Mapping, Dict,
    FrozenSet,
    Sequence, Tuple,
)

from kubernetes_asyncio import client as kube_client, config as kube_config, watch
from aiodocker.docker import Docker
from aiotools import TaskGroup

from ai.backend.common.docker import ImageRef
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.utils import current_loop
import zmq
from ..resources import KernelResourceSpec
from ..kernel import AbstractKernel, AbstractCodeRunner

log = BraceStyleAdapter(logging.getLogger(__name__))


class KubernetesKernel(AbstractKernel):

    # FIXME: apply TypedDict to data in Python 3.8
    deployment_name: str

    def __init__(self, kernel_id: str, image: ImageRef, version: int, *,
                 agent_config: Mapping[str, Any],
                 resource_spec: KernelResourceSpec,
                 service_ports: Any,  # TODO: type-annotation
                 data: Dict[str, Any]) -> None:
        super().__init__(
            kernel_id, image, version,
            agent_config=agent_config,
            resource_spec=resource_spec,
            service_ports=service_ports,
            data=data)

        self.deployment_name = f'kernel-{kernel_id}'

    async def close(self) -> None:
        await self.scale(0)

    async def create_code_runner(self, *,
                           client_features: FrozenSet[str],
                           api_version: int) -> AbstractCodeRunner:

        scale = await self.scale(1)
        if scale.to_dict()['spec']['replicas'] == 0:
            log.error('Scaling failed! Response body: {0}', scale)
            raise ValueError('Scaling failed!')

        if scale.to_dict()['status']['replicas'] == 0:
            while not await self.is_scaled():
                await asyncio.sleep(0.5)

        # TODO: Find way to detect if kernel runner has started inside container

        runner = await KubernetesCodeRunner.new(
            self.kernel_id,
            kernel_host=self.data['kernel_host'],
            repl_in_port=self.data['repl_in_port'],
            repl_out_port=self.data['repl_out_port'],
            exec_timeout=0,
            client_features=client_features)

        retries = 0
        while True:
            try:
                await runner.feed_and_get_status()
                break
            except zmq.error.ZMQError as e:
                if retries < 4:
                    retries += 1
                    log.debug('Socket not responding, retrying #{}', retries)
                    await asyncio.sleep(retries ** 2)
                else:
                    raise e

        return runner

    async def scale(self, num: int):
        await kube_config.load_kube_config()
        apps_api = kube_client.AppsV1Api()
        try:
            return await apps_api.replace_namespaced_deployment_scale(
                self.deployment_name, 'backend-ai',
                body={
                    'apiVersion': 'autoscaling/v1',
                    'kind': 'Scale',
                    'metadata': {
                        'name': self.deployment_name,
                        'namespace': 'backend-ai',
                    },
                    'spec': {'replicas': num},
                    'status': {'replicas': num, 'selector': f'run={self.deployment_name}'},
                },
            )
        except Exception as e:
            log.exception('scale failed: {}', e)

    async def is_scaled(self):
        await kube_config.load_kube_config()
        apps_api = kube_client.AppsV1Api()
        core_api = kube_client.CoreV1Api()
        scale = await apps_api.read_namespaced_deployment(self.deployment_name, 'backend-ai')

        if scale.to_dict()['status']['replicas'] == 0:
            return False
        for condition in scale.to_dict()['status']['conditions']:
            if not condition['status']:
                return False

        pods = await core_api.list_namespaced_pod(
            'backend-ai',
            label_selector=f'run=kernel-{self.kernel_id}',
        )
        pods = pods.to_dict()['items'] or []
        if len(pods) == 0:
            return False
        for pod in pods:
            containers = pod['status']['container_statuses'] or []
            if len(containers) == 0:
                return False
            for container in containers:
                started = container.get('started')
                if not container['ready'] or started is not None and not started:
                    return False
        return True

    async def get_completions(self, text: str, opts: Mapping[str, Any]):
        result = await self.runner.feed_and_get_completion(text, opts)
        return {'status': 'finished', 'completions': result}

    async def check_status(self):
        result = await self.runner.feed_and_get_status()
        return result

    async def get_logs(self):
        await kube_config.load_kube_config()
        core_api = kube_client.CoreV1Api()

        result = await core_api.read_namespaced_pod_log(self.kernel_id, 'backend-ai')
        return {'logs': result.data.decode('utf-8')}

    async def interrupt_kernel(self):
        await self.runner.feed_interrupt()
        return {'status': 'finished'}

    async def start_service(self, service: str, opts: Mapping[str, Any]):
        if self.data.get('block_service_ports', False):
            return {
                'status': 'failed',
                'error': 'operation blocked',
            }
        for sport in self.service_ports:
            if sport['name'] == service:
                break
        else:
            return {'status': 'failed', 'error': 'invalid service name'}
        result = await self.runner.feed_start_service({
            'name': service,
            'port': sport['container_ports'][0],  # primary port
            'ports': sport['container_ports'],
            'protocol': sport['protocol'],
            'options': opts,
        })
        return result

    async def shutdown_service(self, service: str):
        await self.runner.feed_shutdown_service(service)

    async def get_service_apps(self):
        result = await self.runner.feed_service_apps()
        return result

    async def accept_file(self, filename: str, filedata: bytes):
        loop = current_loop()
        work_dir = self.agent_config['container']['scratch-root'] / str(self.kernel_id) / 'work'
        try:
            # create intermediate directories in the path
            dest_path = (work_dir / filename).resolve(strict=False)
            parent_path = dest_path.parent
        except ValueError:  # parent_path does not start with work_dir!
            raise AssertionError('malformed upload filename and path.')

        def _write_to_disk():
            parent_path.mkdir(parents=True, exist_ok=True)
            dest_path.write_bytes(filedata)

        try:
            await loop.run_in_executor(None, _write_to_disk)
        except FileNotFoundError:
            log.error('{0}: writing uploaded file failed: {1} -> {2}',
                      self.kernel_id, filename, dest_path)

    async def download_file(self, filepath: str):
        # TODO: Implement file operations with pure Kubernetes API
        await kube_config.load_kube_config()
        core_api = kube_client.CoreV1Api()

        home_path = Path('/home/work')
        try:
            abspath = (home_path / filepath).resolve()
            abspath.relative_to(home_path)
        except ValueError:
            raise PermissionError('You cannot download files outside /home/work')

        async with watch.Watch().stream(
            core_api.connect_get_namespaced_pod_exec,
            self.kernel_id, 'backend-ai',
            command=['tar', 'cf', '-', abspath.resolve()], stderr=True, stdin=True, stdout=True,
            tty=False, _preload_content=False,
        ) as stream:
            async for event in stream:
                log.debug('stream: {}', event)

        return None

    async def list_files(self, container_path: str):
        # TODO: Implement file operations with pure Kubernetes API
        await kube_config.load_kube_config()
        core_api = kube_client.CoreV1Api()

        # Confine the lookable paths in the home directory
        home_path = Path('/home/work')
        try:
            resolved_path = (home_path / container_path).resolve()
            resolved_path.relative_to(home_path)
        except ValueError:
            raise PermissionError('You cannot list files outside /home/work')

        # Gather individual file information in the target path.
        code = textwrap.dedent('''
        import json
        import os
        import stat
        import sys

        files = []
        for f in os.scandir(sys.argv[1]):
            fstat = f.stat()
            ctime = fstat.st_ctime  # TODO: way to get concrete create time?
            mtime = fstat.st_mtime
            atime = fstat.st_atime
            files.append({
                'mode': stat.filemode(fstat.st_mode),
                'size': fstat.st_size,
                'ctime': ctime,
                'mtime': mtime,
                'atime': atime,
                'filename': f.name,
            })
        print(json.dumps(files))
        ''')

        command = ['/opt/backend.ai/bin/python', '-c', code,  str(container_path)]
        async with watch.Watch().stream(
            core_api.connect_get_namespaced_pod_exec,
            self.kernel_id, 'backend-ai',
            command=command, stderr=True, stdin=True, stdout=True,
            tty=False, _preload_content=False,
        ) as stream:
            async for event in stream:
                log.debug('stream: {}', event)

        return {'files': '', 'errors': '', 'abspath': str(container_path)}


class KubernetesCodeRunner(AbstractCodeRunner):

    kernel_host: str
    repl_in_port: int
    repl_out_port: int

    def __init__(self, kernel_id, *,
                 kernel_host, repl_in_port, repl_out_port,
                 exec_timeout=0, client_features=None) -> None:
        super().__init__(
            kernel_id,
            exec_timeout=exec_timeout,
            client_features=client_features)
        self.kernel_host = kernel_host
        self.repl_in_port = repl_in_port
        self.repl_out_port = repl_out_port

    async def get_repl_in_addr(self) -> str:
        return f'tcp://{self.kernel_host}:{self.repl_in_port}'

    async def get_repl_out_addr(self) -> str:
        return f'tcp://{self.kernel_host}:{self.repl_out_port}'


async def prepare_krunner_env_impl(distro: str, root_path: str) -> Tuple[str, Optional[str]]:
    if distro.startswith('static-'):
        distro_name = distro.replace('-', '_')  # pkg/mod name use underscores
    else:
        if (m := re.search(r'^([a-z]+)\d+\.\d+$', distro)) is None:
            raise ValueError('Unrecognized "distro[version]" format string.')
        distro_name = m.group(1)
    docker = Docker()
    arch = get_arch_name()
    current_version = int(Path(
        pkg_resources.resource_filename(
            f'ai.backend.krunner.{distro_name}',
            f'./krunner-version.{distro}.txt'))
        .read_text().strip())
    krunner_folder_name = f'backendai-krunner.v{current_version}.{distro}'
    target_path = Path(root_path) / krunner_folder_name
    extractor_image = 'backendai-krunner-extractor:latest'

    try:
        for item in (await docker.images.list()):
            if item['RepoTags'] is None:
                continue
            if item['RepoTags'][0] == extractor_image:
                break
        else:
            log.info('preparing the Docker image for krunner extractor...')
            extractor_archive = pkg_resources.resource_filename(
                'ai.backend.runner', f'krunner-extractor.img.{arch}.tar.xz')
            with lzma.open(extractor_archive, 'rb') as reader:
                proc = await asyncio.create_subprocess_exec(
                    *['docker', 'load'], stdin=reader)
                if (await proc.wait() != 0):
                    raise RuntimeError('loading krunner extractor image has failed!')

        log.info('checking krunner-env for {}.{}...', distro, arch)

        if not target_path.exists():
            log.info('populating {} volume version {}',
                     krunner_folder_name, current_version)
            target_path.mkdir(exist_ok=False)
            archive_path = Path(pkg_resources.resource_filename(
                f'ai.backend.krunner.{distro_name}',
                f'krunner-env.{distro}.{arch}.tar.xz')).resolve()
            extractor_path = Path(pkg_resources.resource_filename(
                'ai.backend.runner',
                'krunner-extractor.sh')).resolve()

            log.debug('Executing {}', ' '.join([
                'docker', 'run', '--rm', '-i',
                '-v', f'{archive_path}:/root/archive.tar.xz',
                '-v', f'{extractor_path}:/root/krunner-extractor.sh',
                '-v', f'{target_path.absolute().as_posix()}:/root/volume',
                '-e', f'KRUNNER_VERSION={current_version}',
                extractor_image,
                '/root/krunner-extractor.sh',
            ]))

            proc = await asyncio.create_subprocess_exec(*[
                'docker', 'run', '--rm', '-i',
                '-v', f'{archive_path}:/root/archive.tar.xz',
                '-v', f'{extractor_path}:/root/krunner-extractor.sh',
                '-v', f'{target_path.absolute().as_posix()}:/root/volume',
                '-e', f'KRUNNER_VERSION={current_version}',
                extractor_image,
                '/root/krunner-extractor.sh',
            ])
            if (await proc.wait() != 0):
                raise RuntimeError('extracting krunner environment has failed!')
    except Exception:
        log.exception('unexpected error')
        return distro, None
    finally:
        await docker.close()
    return distro, krunner_folder_name


async def copy_runner_files(scratch_path: Path) -> None:
    artifact_path = Path(pkg_resources.resource_filename('ai.backend.agent', '../runner'))
    kernel_path = Path(pkg_resources.resource_filename('ai.backend.agent', '../kernel'))
    helpers_path = Path(pkg_resources.resource_filename('ai.backend.agent', '../helpers'))

    destination_path = scratch_path

    if (destination_path / 'runner').exists():
        shutil.rmtree(destination_path / 'runner', ignore_errors=True)
    (destination_path / 'runner').mkdir(parents=True)

    target_files = [
        'entrypoint.sh',
        '*.bin',
        '*.so',
        'DO_NOT_STORE_PERSISTENT_FILES_HERE.md',
        'extract_dotfiles.py',
    ]

    for target_glob in target_files:
        for matched_path in artifact_path.glob(target_glob):
            shutil.copy(matched_path.resolve(), destination_path / 'runner')

    if (destination_path / 'kernel').exists():
        shutil.rmtree(destination_path / 'kernel', ignore_errors=True)
    shutil.copytree(kernel_path.resolve(), destination_path / 'kernel')

    if (destination_path / 'helpers').exists():
        shutil.rmtree(destination_path / 'helpers', ignore_errors=True)
    shutil.copytree(helpers_path.resolve(), destination_path / 'helpers')


async def prepare_krunner_env(local_config: Mapping[str, Any]) -> Mapping[str, Sequence[str]]:
    """
    Check if the volume "backendai-krunner.{distro}.{arch}" exists and is up-to-date.
    If not, automatically create it and update its content from the packaged pre-built krunner
    tar archives.
    """

    all_distros = []
    entry_prefix = 'backendai_krunner_v10'
    for entrypoint in pkg_resources.iter_entry_points(entry_prefix):
        log.debug('loading krunner pkg: {}', entrypoint.module_name)
        plugin = entrypoint.load()
        await plugin.init({})  # currently does nothing
        provided_versions = Path(pkg_resources.resource_filename(
            f'ai.backend.krunner.{entrypoint.name}',
            'versions.txt',
        )).read_text().splitlines()
        all_distros.extend(provided_versions)

    scratch_mount = local_config['container']['scratch-root']
    await copy_runner_files(Path(scratch_mount))

    tasks = []
    async with TaskGroup() as tg:
        for distro in all_distros:
            tasks.append(tg.create_task(prepare_krunner_env_impl(distro, scratch_mount)))
    distro_volumes = [t.result() for t in tasks if not t.cancelled()]
    result = {}
    for distro_name_and_version, volume_name in distro_volumes:
        if volume_name is None:
            continue
        result[distro_name_and_version] = volume_name
    return result



import logging
from typing import Any, Mapping
from uuid import UUID
from ai.backend.common.types import KernelId

from aiodocker.docker import Docker
from aiohttp import web

from ai.backend.agent.utils import closing_async
from ai.backend.common.logging import BraceStyleAdapter

from ai.backend.agent.docker.kernel import DockerKernel

from ai.backend.agent.kernel import AbstractKernel
from aiohttp.typedefs import Handler

log = BraceStyleAdapter(logging.getLogger(__name__))


@web.middleware
async def container_resolver(request: web.Request, handler: Handler):
    if request.headers.get('X-Forwarded-For') is not None and request.app['docker-mode'] == 'linuxkit':
        container_ip = request.headers['X-Forwarded-For']
    elif remote_ip := request.remote:
        container_ip = remote_ip
    else:
        return web.Response(status=403)
    async with closing_async(Docker()) as docker:
        containers = await docker.containers.list(
            filters='{"label":["ai.backend.kernel-id"],"network":["bridge"],"status":["running"]}',
        )
    target_container = list(filter(
        lambda x: x['NetworkSettings']['Networks'].get('bridge', {}).get('IPAddress') == container_ip,
        containers,
    ))

    if len(target_container) == 0:
        return web.Response(status=403)
    request['container-ip'] = container_ip
    request['container'] = target_container[0]
    return await handler(request)


async def get_metadata(request: web.Request) -> web.Response:
    kernel: DockerKernel = \
        request.app['kernel-registry'].get(UUID(request['container']['Labels']['ai.backend.kernel-id']))
    if kernel is None:
        return web.Response(status=404)
    response = dict(kernel.environ)
    return web.json_response(response)


# TODO: Split out metadata server as seperate backend.ai plugin
async def create_server(
    local_config: Mapping[str, Any],
    kernel_registry: Mapping[KernelId, AbstractKernel],
) -> web.Application:
    app = web.Application(
        middlewares=[container_resolver],
    )
    app['docker-mode'] = local_config['agent']['docker-mode']
    app['kernel-registry'] = kernel_registry
    app.router.add_route('GET', '/meta-data', get_metadata)
    return app



import logging
import os
from pathlib import Path
from typing import Dict

from ai.backend.common.logging import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__name__))

# the names of following AWS variables follow boto3 convention.
s3_access_key = os.environ.get('AWS_ACCESS_KEY_ID', 'dummy-access-key')
s3_secret_key = os.environ.get('AWS_SECRET_ACCESS_KEY', 'dummy-secret-key')
s3_region = os.environ.get('AWS_REGION', 'ap-northeast-1')
s3_bucket = os.environ.get('AWS_S3_BUCKET', 'codeonweb')
s3_bucket_path = os.environ.get('AWS_S3_BUCKET_PATH', 'bucket')

if s3_access_key == 'dummy-access-key':
    log.info('Automatic ~/.output file S3 uploads is disabled.')


def relpath(path, base):
    return Path(path).resolve().relative_to(Path(base).resolve())


def scandir(root: Path, allowed_max_size: int):
    '''
    Scans a directory recursively and returns a dictionary of all files and
    their last modified time.
    '''
    file_stats: Dict[Path, float] = dict()
    if not isinstance(root, Path):
        root = Path(root)
    if not root.exists():
        return file_stats
    for entry in os.scandir(root):
        # Skip hidden files.
        if entry.name.startswith('.'):
            continue
        if entry.is_file():
            try:
                stat = entry.stat()
            except PermissionError:
                continue
            # Skip too large files!
            if stat.st_size > allowed_max_size:
                continue
            file_stats[Path(entry.path)] = stat.st_mtime
        elif entry.is_dir():
            try:
                file_stats.update(scandir(Path(entry.path), allowed_max_size))
            except PermissionError:
                pass
    return file_stats


def diff_file_stats(fs1, fs2):
    k2 = set(fs2.keys())
    k1 = set(fs1.keys())
    new_files = k2 - k1
    modified_files = set()
    for k in (k2 - new_files):
        if fs1[k] < fs2[k]:
            modified_files.add(k)
    return new_files | modified_files



"""
Parses and interpolates the service-definition templates stored as json in
``/etc/backend.ai/servce-defs`` of Backend.AI containers.

See more details at `the documentation about adding new kernel images
<https://docs.backend.ai/en/latest/dev/adding-kernels.html#service-ports>`_.
"""

import json
import logging
from pathlib import Path
from typing import (
    Any, Collection, Optional, Union,
    TypedDict,
    Mapping, MutableMapping, Dict,
    Sequence, List, Tuple,
)

import attr

from . import service_actions
from .logging import BraceStyleAdapter
from .exception import DisallowedArgument, DisallowedEnvironment, InvalidServiceDefinition

log = BraceStyleAdapter(logging.getLogger())


class Action(TypedDict):
    action: str
    args: Mapping[str, str]
    ref: Optional[str]


@attr.s(auto_attribs=True, slots=True)
class ServiceDefinition:
    command: List[str]
    noop: bool = False
    url_template: str = ''
    prestart_actions: List[Action] = attr.Factory(list)
    env: Mapping[str, str] = attr.Factory(dict)
    allowed_envs: List[str] = attr.Factory(list)
    allowed_arguments: List[str] = attr.Factory(list)
    default_arguments: Mapping[str, Union[None, str, List[str]]] = attr.Factory(dict)


class ServiceParser:

    variables: MutableMapping[str, str]
    services: MutableMapping[str, ServiceDefinition]

    def __init__(self, variables: MutableMapping[str, str]) -> None:
        self.variables = variables
        self.services = {}

    async def parse(self, path: Path) -> None:
        for service_def_file in path.glob('*.json'):
            log.debug(f'loading service-definition from {service_def_file}')
            try:
                with open(service_def_file.absolute(), 'rb') as fr:
                    raw_service_def = json.load(fr)
                    # translate naming differences
                    if 'prestart' in raw_service_def:
                        raw_service_def['prestart_actions'] = raw_service_def['prestart']
                        del raw_service_def['prestart']
            except IOError:
                raise InvalidServiceDefinition(
                    f'could not read the service-def file: {service_def_file.name}')
            except json.JSONDecodeError:
                raise InvalidServiceDefinition(
                    f'malformed JSON in service-def file: {service_def_file.name}')
            name = service_def_file.stem
            try:
                self.services[name] = ServiceDefinition(**raw_service_def)
            except TypeError as e:
                raise InvalidServiceDefinition(e.args[0][11:])  # lstrip "__init__() "

    async def start_service(
        self,
        service_name: str,
        frozen_envs: Collection[str],
        opts: Mapping[str, Any],
    ) -> Tuple[Optional[Sequence[str]], Mapping[str, str]]:
        if service_name not in self.services.keys():
            return None, {}
        service = self.services[service_name]
        if service.noop:
            return [], {}

        for action in service.prestart_actions:
            try:
                action_impl = getattr(service_actions, action['action'])
            except AttributeError:
                raise InvalidServiceDefinition(
                    f"Service-def for {service_name} used invalid action: {action['action']}")
            ret = await action_impl(self.variables, **action['args'])
            if (ref := action.get('ref')) is not None:
                self.variables[ref] = ret

        cmdargs, env = [], {}

        for arg in service.command:
            cmdargs.append(arg.format_map(self.variables))

        additional_arguments = dict(service.default_arguments)
        if 'arguments' in opts.keys():
            for argname, argvalue in opts['arguments'].items():
                if argname not in service.allowed_arguments:
                    raise DisallowedArgument(
                        f'Argument {argname} not allowed for service {service_name}')
                additional_arguments[argname] = argvalue

        for env_name, env_value in service.env.items():
            env[env_name.format_map(self.variables)] = env_value.format_map(self.variables)

        if 'envs' in opts.keys():
            for envname, envvalue in opts['envs'].items():
                if envname not in service.allowed_envs:
                    raise DisallowedEnvironment(
                        f'Environment variable {envname} not allowed for service {service_name}')
                elif envname in frozen_envs:
                    raise DisallowedEnvironment(
                        f'Environment variable {envname} can\'t be overwritten')
                env[envname] = envvalue

        for arg_name, arg_value in additional_arguments.items():
            cmdargs.append(arg_name)
            if isinstance(arg_value, str):
                cmdargs.append(arg_value)
            elif isinstance(arg_value, list):
                cmdargs += arg_value

        return cmdargs, env

    async def get_apps(self, selected_service: str = '') -> Sequence[Mapping[str, Any]]:

        def _format(service_name: str) -> Mapping[str, Any]:
            service_info: Dict[str, Any] = {'name': service_name}
            service = self.services[service_name]
            if len(service.url_template) > 0:
                service_info['url_template'] = service.url_template
            if len(service.allowed_arguments) > 0:
                service_info['allowed_arguments'] = service.allowed_arguments
            if len(service.allowed_envs) > 0:
                service_info['allowed_envs'] = service.allowed_envs
            return service_info

        apps = []
        if selected_service:
            if selected_service in self.services.keys():
                apps.append(_format(selected_service))
        else:
            for service_name in self.services.keys():
                apps.append(_format(service_name))
        return apps



from decimal import Decimal
import logging
import os
from pathlib import Path
import platform
from typing import (
    Any,
    Collection,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
)

import aiohttp
from aiodocker.docker import Docker, DockerContainer
from aiodocker.exceptions import DockerError
from kubernetes_asyncio import client as K8sClient, config as K8sConfig

from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import (
    DeviceName, DeviceId,
    DeviceModelInfo,
    SlotName, SlotTypes,
)
from .agent import Container
from .resources import (
    get_resource_spec_from_container,
)
from .. import __version__
from ..resources import (
    AbstractAllocMap, DeviceSlotInfo,
    DiscretePropertyAllocMap,
    AbstractComputeDevice,
    AbstractComputePlugin,
)
from ..stats import (
    StatContext, NodeMeasurement, ContainerMeasurement,
)

log = BraceStyleAdapter(logging.getLogger(__name__))


async def fetch_api_stats(container: DockerContainer) -> Optional[Dict[str, Any]]:
    short_cid = container._id[:7]
    try:
        ret = await container.stats(stream=False)  # TODO: cache
    except RuntimeError as e:
        msg = str(e.args[0]).lower()
        if 'event loop is closed' in msg or 'session is closed' in msg:
            return None
        raise
    except (DockerError, aiohttp.ClientError) as e:
        log.error(
            'cannot read stats (cid:{}): client error: {!r}.',
            short_cid, e,
        )
        return None
    else:
        # aiodocker 0.16 or later returns a list of dict, even when not streaming.
        if isinstance(ret, list):
            if not ret:
                # The API may return an empty result upon container termination.
                return None
            ret = ret[0]
        # The API may return an invalid or empty result upon container termination.
        if ret is None or not isinstance(ret, dict):
            log.warning(
                'cannot read stats (cid:{}): got an empty result: {}',
                short_cid, ret,
            )
            return None
        if (
            ret['read'].startswith('0001-01-01') or
            ret['preread'].startswith('0001-01-01')
        ):
            return None
        return ret


# Pseudo-plugins for intrinsic devices (CPU and the main memory)

class CPUDevice(AbstractComputeDevice):
    pass


class CPUPlugin(AbstractComputePlugin):
    """
    Represents the CPU.
    """

    config_watch_enabled = False

    key = DeviceName('cpu')
    slot_types = [
        (SlotName('cpu'), SlotTypes.COUNT),
    ]

    async def init(self, context: Any = None) -> None:
        pass

    async def cleanup(self) -> None:
        pass

    async def update_plugin_config(self, new_plugin_config: Mapping[str, Any]) -> None:
        pass

    async def list_devices(self) -> Collection[CPUDevice]:
        await K8sConfig.load_kube_config()
        core_api = K8sClient.CoreV1Api()

        nodes = (await core_api.list_node()).to_dict()['items']
        overcommit_factor = int(os.environ.get('BACKEND_CPU_OVERCOMMIT_FACTOR', '1'))
        assert 1 <= overcommit_factor <= 10

        return [
            CPUDevice(
                device_id=DeviceId(node['metadata']['uid']),
                hw_location='root',
                numa_node=None,
                memory_size=0,
                processing_units=int(node['status']['capacity']['cpu']) * overcommit_factor,
            )
            for i, node in zip(range(len(nodes)), nodes)
            # if 'node-role.kubernetes.io/master' not in node['metadata']['labels'].keys()
        ]

    async def available_slots(self) -> Mapping[SlotName, Decimal]:
        devices = await self.list_devices()
        log.debug('available_slots: {}', devices)
        return {
            SlotName('cpu'): Decimal(sum(dev.processing_units for dev in devices)),
        }

    def get_version(self) -> str:
        return __version__

    async def extra_info(self) -> Mapping[str, str]:
        return {
            'agent_version': __version__,
            'machine': platform.machine(),
            'os_type': platform.system(),
        }

    async def gather_node_measures(self, ctx: StatContext) -> Sequence[NodeMeasurement]:
        # TODO: Create our own k8s metric collector

        return []

    async def gather_container_measures(
        self,
        ctx: StatContext,
        container_ids: Sequence[str],
    ) -> Sequence[ContainerMeasurement]:
        # TODO: Implement Kubernetes-specific container metric collection

        return [
        ]

    async def create_alloc_map(self) -> AbstractAllocMap:
        devices = await self.list_devices()
        return DiscretePropertyAllocMap(
            device_slots={
                dev.device_id:
                    DeviceSlotInfo(SlotTypes.COUNT, SlotName('cpu'), Decimal(dev.processing_units))
                for dev in devices
            },
        )

    async def get_hooks(self, distro: str, arch: str) -> Sequence[Path]:
        # TODO: move the sysconf hook in libbaihook.so here
        return []

    async def generate_docker_args(
        self,
        docker: Docker,
        device_alloc,
    ) -> Mapping[str, Any]:
        # This function might be needed later to apply fine-grained tuning for
        # K8s resource allocation
        return {}

    async def restore_from_container(
        self,
        container: Container,
        alloc_map: AbstractAllocMap,
    ) -> None:
        assert isinstance(alloc_map, DiscretePropertyAllocMap)
        # Docker does not return the original cpuset.... :(
        # We need to read our own records.
        resource_spec = await get_resource_spec_from_container(container.backend_obj)
        if resource_spec is None:
            return
        alloc_map.apply_allocation({
            SlotName('cpu'):
                resource_spec.allocations[DeviceName('cpu')][SlotName('cpu')],
        })

    async def get_attached_devices(
        self,
        device_alloc: Mapping[SlotName,
        Mapping[DeviceId, Decimal]],
    ) -> Sequence[DeviceModelInfo]:
        device_ids = [*device_alloc[SlotName('cpu')].keys()]
        available_devices = await self.list_devices()
        attached_devices: List[DeviceModelInfo] = []
        for device in available_devices:
            if device.device_id in device_ids:
                attached_devices.append({
                    'device_id': device.device_id,
                    'model_name': '',
                    'data': {'cores': len(device_ids)},
                })
        return attached_devices


class MemoryDevice(AbstractComputeDevice):
    pass


class MemoryPlugin(AbstractComputePlugin):
    """
    Represents the main memory.

    When collecting statistics, it also measures network and I/O usage
    in addition to the memory usage.
    """

    config_watch_enabled = False

    key = DeviceName('mem')
    slot_types = [
        (SlotName('mem'), SlotTypes.BYTES),
    ]

    async def init(self, context: Any = None) -> None:
        pass

    async def cleanup(self) -> None:
        pass

    async def update_plugin_config(self, new_plugin_config: Mapping[str, Any]) -> None:
        pass

    async def list_devices(self) -> Collection[MemoryDevice]:
        await K8sConfig.load_kube_config()
        core_api = K8sClient.CoreV1Api()

        nodes = (await core_api.list_node()).to_dict()['items']
        overcommit_factor = int(os.environ.get('BACKEND_MEM_OVERCOMMIT_FACTOR', '1'))
        assert 1 <= overcommit_factor <= 10
        mem = 0
        for node in nodes:
            # if 'node-role.kubernetes.io/master' in node['metadata']['labels'].keys():
            #     continue
            mem += int(node['status']['capacity']['memory'][:-2]) * 1024
        return [
            MemoryDevice(
                device_id=DeviceId('root'),
                hw_location='root',
                numa_node=0,
                memory_size=mem * overcommit_factor,
                processing_units=0,
            ),
        ]

    async def available_slots(self) -> Mapping[SlotName, Decimal]:
        devices = await self.list_devices()
        return {
            SlotName('mem'): Decimal(sum(dev.memory_size for dev in devices)),
        }

    def get_version(self) -> str:
        return __version__

    async def extra_info(self) -> Mapping[str, str]:
        return {}

    async def gather_node_measures(self, ctx: StatContext) -> Sequence[NodeMeasurement]:
        # TODO: Create our own k8s metric collector
        return []

    async def gather_container_measures(self, ctx: StatContext, container_ids: Sequence[str]) \
            -> Sequence[ContainerMeasurement]:
        # TODO: Implement Kubernetes-specific container metric collection
        return []

    async def create_alloc_map(self) -> AbstractAllocMap:
        devices = await self.list_devices()
        return DiscretePropertyAllocMap(
            device_slots={
                dev.device_id:
                    DeviceSlotInfo(SlotTypes.BYTES, SlotName('mem'), Decimal(dev.memory_size))
                for dev in devices
            },
        )

    async def get_hooks(self, distro: str, arch: str) -> Sequence[Path]:
        return []

    async def generate_docker_args(
        self,
        docker: Docker,
        device_alloc,
    ) -> Mapping[str, Any]:
        # This function might be needed later to apply fine-grained tuning for
        # K8s resource allocation
        return {}

    async def restore_from_container(
        self,
        container: Container,
        alloc_map: AbstractAllocMap,
    ) -> None:
        assert isinstance(alloc_map, DiscretePropertyAllocMap)
        memory_limit = container.backend_obj['HostConfig']['Memory']
        alloc_map.apply_allocation({
            SlotName('mem'): {DeviceId('root'): memory_limit},
        })

    async def get_attached_devices(
        self,
        device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]],
    ) -> Sequence[DeviceModelInfo]:
        device_ids = [*device_alloc[SlotName('mem')].keys()]
        available_devices = await self.list_devices()
        attached_devices: List[DeviceModelInfo] = []
        for device in available_devices:
            if device.device_id in device_ids:
                attached_devices.append({
                    'device_id': device.device_id,
                    'model_name': '',
                    'data': {},
                })
        return attached_devices
